/* Embedded OS Loader
 *
 *   Copyright 2005 Vivien Chappelier <vivien.chappelier@free.fr>
 *                  Thomas Cougnard <thomas.cougnard@free.fr>
 *
 * bootloader
 *
 */

#include "stdlib.h"
#include "fb.h"
#include "mmc.h"
#include "disk.h"
#include "fat.h"
#include "tag.h"

extern void enter_linux(void *address, unsigned int mach_type, void *taglist);
extern void dead_start();
extern void dead_end();

extern void *end;
extern void *start;

#define IMAGE_NAME     "IMAGE.BIN"
#define INITRD_NAME    "INITRD.BIN"
#define LINUX_VIRT     0x8c008000 /* virtual address where linux is loaded */
#define LINUX_PHYS     0x10008000 /* physical address where linux is loaded */
#define TAGLIST_VIRT   0x8c000100 /* virtual address of boot parameters */
#define TAGLIST_PHYS   0x10000100 /* physical address of boot parameters */
#define TAGLIST_SIZE   0x00001000 /* this should be largely enough */
#define INITRD_PHYS    0x10800000 /* physical address of initrd (8Mo) */

/* from http://www.arm.linux.org.uk/developer/machines */
#define MACH_TYPHOON 689 


int main(void)
{
  int i;
  int found = 0;
  unsigned char buffer[512*32];
  pentry_t ptable[4];
  fat_t fat;
  fat_dir_entry_t *entry;
  cluster_t fd;
  int size;

  /* initialize the console */
  console_init();

  printb("EOL ");
  printb("%p - %p\n", start, end);

  /* copy safe code at LINUX_VIRT in case something goes wrong */
  memcpy((void *) LINUX_VIRT, dead_start, dead_end - dead_start);

  /* read the SD card */
  mmc_reset();
  mmc_read(0, buffer, 512);

  /* load the partition table */
  read_ptable(buffer, ptable);
  printb("SD card partition table:\n");
  print_pentry(&ptable[0]);
  print_pentry(&ptable[1]);
  print_pentry(&ptable[2]);
  print_pentry(&ptable[3]);

  /* look for the first valid FAT partition */
  for(i = 0; i < 4; i++) {
    switch(ptable[i].type) {
    case 0x00: /* empty */
      break;
    case 0x01: /* FAT12 */
    case 0x04: /* FAT16 <32M */
    case 0x06: /* FAT16 */
    case 0x0b: /* FAT32 */
    case 0x0c: /* FAT32 (LBA) */
    case 0x0e: /* FAT16 (LBA) */
      found = 1;
      break;
    default:
      printb("partition %x: type %x unknown\n", i, ptable[i].type);
      break;
    }
    if(found) break;
  }
  if(!found) {
    printb("No valid partition found\n");
    return(1);
  }

  /* read the FAT */
  if(!fat_init(&fat, ptable[i].sector_start)) {
    printb("FAT init failure\n");
    return(1);
  }
  
  printb("FAT%d %s %dK\n", fat.type, fat.label,
	     fat.sectors_per_cluster * 512 / 1024);

  /* list the files in the FAT root directory */
  for(entry = fat.root; entry; entry = entry->next)
    printb(" %s %d 0x%x\n", entry->filename, entry->filesize, entry->cluster);
  
  /* load linux */
  fd = fat_open_file(&fat, IMAGE_NAME);
  if(!fd) {
         printb("Could not open %s\n", IMAGE_NAME);
         die();
  }

  size = fat_read_file(&fat, fd, (void *) LINUX_VIRT);
  printb("Kernel loaded... (%dKb)\n", size / 1024);
  /*
  printb("%x %x %x %x %x %x %x %x\n",
	 ((unsigned char *) LINUX_LOADADDR)[0],
	 ((unsigned char *) LINUX_LOADADDR)[1],
	 ((unsigned char *) LINUX_LOADADDR)[2],
	 ((unsigned char *) LINUX_LOADADDR)[3],
	 ((unsigned char *) LINUX_LOADADDR)[4],
	 ((unsigned char *) LINUX_LOADADDR)[5],
	 ((unsigned char *) LINUX_LOADADDR)[6],
	 ((unsigned char *) LINUX_LOADADDR)[7]);
  printb("%x %x %x %x %x %x %x %x\n",
	 ((unsigned char *) LINUX_LOADADDR)[8],
	 ((unsigned char *) LINUX_LOADADDR)[9],
	 ((unsigned char *) LINUX_LOADADDR)[10],
	 ((unsigned char *) LINUX_LOADADDR)[11],
	 ((unsigned char *) LINUX_LOADADDR)[12],
	 ((unsigned char *) LINUX_LOADADDR)[13],
	 ((unsigned char *) LINUX_LOADADDR)[14],
	 ((unsigned char *) LINUX_LOADADDR)[15]);
  */

  /* load initrd */
  size = 0;
  fd = fat_open_file(&fat, INITRD_NAME);
  if(fd) {
	size = fat_read_file(&fat, fd, (void *) INITRD_PHYS);

 	printb("Ramdisk loaded... (%dKb)\n", size / 1024);
  }

  /* prepare the tag list (parameters for linux) */
  prepare_tag_list((void *) TAGLIST_VIRT, (void *) INITRD_PHYS, size);

  printb("Entering linux... (%dKb)\n", size / 1024);

  /* wait for 1 sec */
  udelay(1000000);

  /* turn off the lcd */
  lcd_off();

  /* enter the gate, Charon awaits */
  enter_linux((void *) LINUX_PHYS, MACH_TYPHOON, (void *) TAGLIST_PHYS);

  return(0);
}
