/* Embedded OS Loader
 *
 *   Copyright 2005 Vivien Chappelier <vivien.chappelier@free.fr>
 *
 * standard routines
 *
 */

#include <stdarg.h>
#include "stdlib.h"
#include "fb.h"

int timer(void) /* decreases */
{
  volatile int *clock = (int *) OMAP730_TIMER_BASE;
  return(*clock);
}

void udelay(int timeout)
{
  int start;

  start = timer();
  while(start - timer() <= ((timeout * 7) >> 3) + 1);
}

void die(void)
{
  while(1);
}

extern void *malloc_base;
void *malloc(size_t size)
{
  void *ptr = malloc_base;
  malloc_base += (size + 3) & ~3; /* keep aligned */
  //printb("malloc: 0x%x\n", ptr);
  return(ptr);
}

void *memset(void *buffer, int val, size_t length)
{
  unsigned char *mem = (unsigned char *) buffer;
  int i;
  
  for(i = 0; i < length; i++)
    mem[i] = (unsigned char) (val & 0xff);
  return(buffer);
}

void *memcpy(void *output, const void *input, size_t length)
{
  const unsigned char *in = (const unsigned char *) input;
  unsigned char *out = (unsigned char *) output;
  int i;
  
  for(i = 0; i < length; i++)
    out[i] = in[i];
  return(output);
}

size_t strlen(const char *str)
{
  int i;

  for(i = 0; *str++; i++);
  return(i);
}

char const *strchr(char const *str, int c)
{
  while(*str) {
    if(*str == c) return(str);
    str++;
  }
  return(NULL);
}

char const *strrchr(char const *str, int c)
{
  char const *last = str + strlen(str) - 1;

  while(last >= str) {
    if(*last == c) return(last);
    last--;
  }
  return(NULL);
}

int __strncmp(const char *str1, const char *str2, size_t count)
{
  int i;
  
  for(i = 0; i < count; i++) {
    if(*str1 < *str2) return(-1);
    if(*str1 > *str2) return(+1);
    if(*str1 == 0) return(0);
    str1++;
    str2++;
  }
  return(0);
}

int __div(int a, int b)
{
  int i = 0;
  int s = 1;

  if(a < 0) {
    a = -a;
    s = -s;
  }
  if(b < 0) {
    b = -b;
    s = -s;
  }

  while(a >= b) {
    a -= b;
    i++;
  }

  return(s*i);
}

int puts(const char *string)
{
  while(*string) putchar(*string++);
  putchar('\n');
  return(0);
}

static const int dec[] = { 1000000000, 100000000, 10000000, 1000000, 100000, 10000, 1000, 100, 10, 1, 0 };

void printb(const char *format, ...)
{
  va_list ap;
  int d;
  int i;
  char c, *str;
  int f;

  va_start(ap, format);
  while(1) {
    f = *format++;
    if(!f) break;

    if(f == '%') { /* special char */
      f = *format++;
      if(!f) break;

      switch(f) {
      case 's':           /* string */
	str = va_arg(ap, char *);
	while(*str) putchar(*str++);
	break;
      case 'c':           /* char */
	c = (char) va_arg(ap, int);
	putchar(c);
	break;
      case 'd':           /* int */
	d = va_arg(ap, int);
	if(!d) {
	  putchar('0');
	  break;
	}
	for(i = 0; d < dec[i]; i++);
	while(dec[i]) {
	  int r = __div(d, dec[i]);
	  putchar('0' + r);
	  d -= r * dec[i];
	  i++;
	}
	break;
      case 'x':           /* int (hexa) */
	d = va_arg(ap, int);
	if(!d) {
	  putchar('0');
	  break;
	}
	i = 28;
	while(!(d >> i)) i -= 4;
	while(i >= 0) {
	  if(((d >> i) & 0xf) > 9)
	    putchar('a' + ((d >> i) & 0xf) - 10);
	  else
	    putchar('0' + ((d >> i) & 0xf));
	  i -= 4;
	}
	break;
      case 'p':           /* pointer */
	d = (int) va_arg(ap, void *); /* 32 bit */
	putchar('0');
	putchar('x');
	i = 28;
	while(i >= 0) {
	  if(((d >> i) & 0xf) > 9)
	    putchar('a' + ((d >> i) & 0xf) - 10);
	  else
	    putchar('0' + ((d >> i) & 0xf));
	  i -= 4;
	}
	break;
      default:            /* unknown format */
	putchar('?');
	break;
      }
    } else {       /* standard char */
      putchar(f);
    }
  }

  va_end(ap);
}

